//
//  DvTime.h
//
//  Dvos - time and time interval object classes
//
//  Tony Allen
//  November 2012
//  January 2015 - added TT2000
//  2016 - leap second handling
//  2017 - change in method of leap second handling 

// TAI and TT are regular time counters with fixed seconds based on atomic clocks
// They both count from noon.
// UTC has leap seconds added to keep clocks (broken down time and time strings)
// within 0.9 sec of the apparent time on Earth (UT1)
// Time strings and hence broken down time are ALWAYS given in UTC
// but data are stored with leap seconds (DV = TAI + 12h ) to allow robust arithmetic.
// UTC strings can have a minute with 61 seconds ( 0 - 60) with the 
// extra second inserted at the end of a day.
// CDF epoch and epoch16 are not well defined as they assume that times are input in UTC
// but do not retain the leap second (which is overlayed on the next second)
// Some missions do not add leap seconds and their (apparent) clocks in UTC will drift due to accumulated leap seconds.
// Cluster adds Leap seconds, so only the actual leap second is lost, but data files break at the end of day
// so this will rarely be visible.

// Values refered to as Epoch (UT) are seconds from MJD2000 without leap seconds.
// It is effectively UTC without leap seconds, so breaking down into UTC year, month day etc is easier
// with the exception of the actual leap second. UT and UTC are here taken to measure from the start of the day.
// They are retained for backward compatibility and convenience for breaking down into UTC strings
// As actual leap seconds appear overlayed with next second, a flag must be used to identify when the seconds field is 60.

// Arithmetic is always done in nanoseconds from start of MJD base day, and is therefore robust.
// When converting between UTC and TAI (or TT) the leap seconds offset may be thought of as moving
// the base_day into the correct time system.

#ifndef DV_TIME_H
#define DV_TIME_H

#include <string.h>
#include <stdlib.h>
#include <cmath>
#include <vector>
#include <iostream>
#include <iomanip>
#include <stdlib.h>
#include <time.h>

using namespace std;

class DvString;

typedef long long DvInt8;

// conversion from Nano-sec to sec
#define Sec2Nano 1000000000

// MJD for 2000-01-01 used as QSAS default base in GUI feedback
#define MJD2000 51544

// MJD for 0000-01-01 (correctly Jan 01, BCE 1) 
// Julian proleptic calendar value. 
#define MJD_0000J -678943

// Note Epoch (UT) do not use leap seconds.
// UTC have leap seconds in the string representation and is thus not a uniform counter
// TAI and TT2000 do include leap seconds when converting to ISO strings

// NASA CDF TT2000 offset in seconds = 43200 - 32.184
// Terrestrial Time == TAI + 32.184
#define TT2000offset 43167816000000
#define TAItoTT         32184000000


// Atomic clock (TAI) offset in nanoseconds (12 hrs)
#define NoonOffset 43200000000000

/* Gregorian proleptic calendar value.  (At MJD_0000J the Gregorian proleptic
   calendar reads two days behind the Julian proleptic calendar, i.e. - 2 days,
   see http://en.wikipedia.org/wiki/Proleptic_Gregorian_calendar,
   so MJD_0000G = MJD_0000J+2) */
#define MJD_0000G -678941

// MJD for Jan 01, 1970 00:00:00 used for seconds from Unix base epoch
#define MJD_1970 40587

// QSAS version number for using DvTime
#define QMJD_TIME_VERSION_START 20400

#define DEFAULT_DECIMAL_SEC 3
#define DEFAULT_GREGORIAN_START "1582-10-15"

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ class DvTime

// Date and time
// MJD measures from the start of 17 Nov 1858 */
	
//	 Default uses the Gregorian calendar after 4 Oct 1582 (Julian) i.e. from 15 Oct 1582 Gregorian
//	 Note unix libraries use Gregorian only from 14 Sept 1752 
//	 More detailed discussion can be found at http://aa.usno.navy.mil/data/docs/JulianDate.php
//	 These routines have been compared with the results of the US Naval Observatory online converter.
//	 Modified Julian Date (MJD) = Julian Date (JD) - 2400000.5
	 
//	 In all routines, specifying a day, hour, minute or second field greater than would be valid is
//	 handled with modulo arithmetic and safe. 
//	 Thus 2006-12-32 00:62:00.0 will safely, and correctly, be treated as 2007-01-01 01:02:00.0 

//-------------------------------------------------------------------------
class DvTime
{
  	
public:

// Date and time are held in Modified Julian Date split into int a day (MJD day) and 
// nanoseconds from start of that day. Nanoseconds may exceed a day or be negative
// DvInt8 has 64 bits and stores ~9x10^9, which gives 10^5 days an chosen nanosec resolution
// Adding offsets of around 300 years or more will fail unless done incrementally 
// and the result normalized each step.

	int base_day; // integer part of MJD used as default. Days are of fixed length, no leap seconds
	DvInt8 time_nsec; /* nano-seconds from start of base_day, effectively holds all leap seconds */
    
// Constructors are in UTC

	DvTime(double sec); 	//  day is Jan 01, 2000 = MJD 51544
	DvTime(DvInt8 nano=0, int day=MJD2000); 	// default day is Jan 01, 2000 = MJD 51544
	DvTime( const char* iso_time);
	DvTime( const DvString& iso_time);
	DvTime( int year, int month, int day, int hour, int min, double sec);
	DvTime( int year, int month=1, int day=1, int hour=0, int min=0, DvInt8 nano=0);
	DvTime( const DvTime& ttime );
	
// Methods
	int  setSecResolution(int resolution); // returns previous resolution
	int  getSecResolution(); // returns previous resolution
	void setFromUT(int year, int month, int day, int hour, int min, double sec);
	void setFromUT(int year, int month, int day, int hour, int min, DvInt8 nano);
	void setFromDOY(int year, int doy, int hour, int min, double sec);
	void setFromBCE(int yearBCE, int month, int day, int hour, int min, double sec);
	void setFromMJD(double ModifiedJulianDate);
	void setFromCDFepoch(double cdfepoch);
	void setFromCDFepoch16(double *cdfepoch16);
	void setFromJD(double JulianDate);
    void setFromEpoch2000(double sec); // deprecated
    void setFromTAI2000(DvInt8 sec); 
	void setFromTT2000(DvInt8 nsec);
	void setFromEpoch1970(double epoch1970); // deprecated
	bool setFromISOstring(const char* ISOstring);
	bool setFromISOstring(DvString ISOstring);

	int breakDownTAI(int *year, int *month, int *day, int *hour, int *min, int *sec, int *nano) const;
	int breakDownUTC(int *year, int *month, int *day, int *hour, int *min, int *sec, int *nano) const;
	DvInt8 getTT2000();
	DvInt8 getTAI2000();
	double getSecTAI();
	double getMJD() const;
	double getJD(); 
	double getDiffDays(const DvTime &ttime)const;
	double getDiffSecs(const DvTime &ttime)const;
	DvInt8 getDiffNano(const DvTime &ttime)const;
	double getCDFepoch();
	void getCDFepoch16(double *epoch16);
	double getEpoch2000(); // deprecated
	double getEpoch1970(); // deprecated 
    DvString getISOstring(int delim=0) const;
	const char * getDayOfWeek();
	const char * getLongDayOfWeek();
	const char * getMonth(int m);
	const char * getLongMonth(int m);
	bool getYandD(int *year, int* day);
	size_t strfUTC(char * buf, size_t len, const char *format);
    void normalize();
    DvTime normalized() const;
    DvInt8 UTtoDV();
    DvInt8 UTtoDV(int y, int m, int d);
    DvInt8 DVtoUT() const;  
    DvInt8 DVtoUT(bool &inLeapSec) const;  
	void setGregorianStartMJD(double GregorianMJD);

 
	DvTime& operator=( const DvTime& ttime )
	{
		// assign from another Time
		base_day = ttime.base_day;
		time_nsec = ttime.time_nsec;

		return *this;
	}
   
	DvTime& operator=( const double dtime )
	{
		// assign from another Time
		*this = DvTime(dtime);

		return *this;
	}
   
	DvTime& operator=( const string& isotime )
	{
		// set from string in ISO time format
		if( !setFromISOstring(isotime.c_str()) ) 
		{
			base_day = MJD2000;
			time_nsec = 0;
		}
		
		return *this;
	}
  
    DvTime& aveThis(const DvTime& ttime);
    
    DvTime ave(const DvTime& ttime)const{
        
        DvTime tres(*this);
        
		return tres.aveThis(ttime);

    }
    
   
    // -----------------------------
 	double operator-(DvTime &t) 
	{
		return this->getDiffSecs(t);
	} 
  
	// --- add and subtract seconds 
	DvTime operator+(const double seconds) const
	{
       DvTime tres(*this);
       tres.time_nsec += (DvInt8)((double)Sec2Nano * seconds);
              
       return tres;
	} 
	DvTime operator+(const DvInt8 nano) const
	{
       DvTime tres=*this;
       tres.time_nsec += nano;
       return tres;
	} 
      
	DvTime operator-(const double seconds) const
	{
       DvTime tres=*this;
       tres.time_nsec -= (DvInt8)((double)Sec2Nano * seconds);
       return tres;
	}
	DvTime operator-(const DvInt8 nano) const
	{
       DvTime tres=*this;
       tres.time_nsec -= nano;
       return tres;
	}
       
	// --- add and subtract time and update
	DvTime& operator+=(const double seconds)
	{
       time_nsec += (DvInt8)((double)Sec2Nano * seconds);
       return *this;
	}
	DvTime& operator+=(const DvInt8 nano)
	{
       time_nsec += nano;
       return *this;
	}
   
	DvTime& operator-=(const double seconds)
	{
       time_nsec -= (DvInt8)((double)Sec2Nano * seconds);;
       return *this;
	}
	DvTime& operator-=(const DvInt8 nano)
	{
       time_nsec -= nano;
       return *this;
	}

/*   
	double operator=(const DvTime& ttime) const
	{
		return time_nsec + base_day * 86400.;
	}
*/   
	// --- logical operators
	bool operator==(const DvTime& ttime) const
	{ 
		if(time_nsec == ttime.time_nsec &&  base_day == ttime.base_day ) return true;
        
        // try forcing same day
        DvTime t1 = this->normalized();
        DvTime t2 = ttime.normalized();
        
        if(t1.time_nsec == t2.time_nsec &&  t1.base_day == t2.base_day) return true;

        return false;
	} 
   
	bool operator!=(const DvTime& ttime) const
	{ 
        DvTime t1 = this->normalized();
        DvTime t2 = ttime.normalized();
        
        if( t1.time_nsec != t2.time_nsec ||  t1.base_day != t2.base_day) return true;
        
        return false;
	} 
   
	bool operator<(const DvTime& ttime) const
	{ 
   		if( this->getDiffNano(ttime) < 0) return true;
        return false;
	}
   
	bool operator<=(const DvTime& ttime) const
	{ 
        if( this->getDiffNano(ttime) <= 0) return true;

        return false;
	}  
   
	bool operator>(const DvTime& ttime) const
	{ 
   		if( this->getDiffNano(ttime) > 0) return true;
        return false;
	}
   
	bool operator>=(const DvTime& ttime) const
	{ 
        if( this->getDiffNano(ttime) >= 0) return true;
        return false;
	}  
   
   DvTime& set_today()
   {
     time_t ltime;
     time(&ltime);
     tm *t=localtime(&ltime);
     setFromUT(t->tm_year+1900, t->tm_mon+1, t->tm_mday, 0, 0, 0.0);
     return *this;
   }

};

// ----------- End of class DvTime -------------------

    

// ++++++++++++++++++++++++++++++++++++++++++++++++++++ class DvEvent

class DvEvent 
{
  
public:
    typedef enum Sense_e {TIME_SENSE_NEGATIVE, NO_TIME_SENSE, TIME_SENSE_POSITIVE} Sense;

private:
    DvTime _t1;
    DvTime _t2;

public:

    // --- Constructors
    DvEvent():_t1(DvTime()),_t2(DvTime()){_t2 += (DvInt8)1;}
    DvEvent( const DvEvent& tint): _t1(tint._t1),_t2(tint._t2){;}
    DvEvent( const DvTime& st, const DvTime& ed, Sense option=NO_TIME_SENSE );
    DvEvent( double st, double ed, Sense option=NO_TIME_SENSE ): _t1(DvTime(st)),_t2(DvTime(ed)){;}
    DvEvent( string &ISOstring );
	    
    // --- get start/end 
    
    DvTime start() const { return _t1;}
    DvTime end() const { return _t2;}
    DvTime start_abs() const { 
		if(_t1<_t2) return _t1;
		else return _t2;
	}
    DvTime end_abs() const { 
		if (_t1>_t2) return _t1;
		else return _t2;
	}  
	
	DvString getISOstring(int delim=0) const;

    // --- get duration 
    
    double duration() const { return _t2.getDiffSecs(_t1);}
    double duration_abs() const {return fabs(_t2.getDiffSecs(_t1));}
    DvInt8 durationNano() const { return _t2.getDiffNano(_t1);}
    DvInt8 durationNano_abs() const {return fabs(_t2.getDiffNano(_t1));}
    DvTime centre() const {
		DvTime tc = _t1.ave(_t2);
		return tc;
	}
         
   
    // --- get sense of time direction:  
    // +1 if start<end, -1 if end<start, 0 if start==end
    Sense get_sense() const
    {
      if (_t1==_t2)
		return NO_TIME_SENSE; 
      else if(_t1>_t2)
        return TIME_SENSE_NEGATIVE;
      else
        return TIME_SENSE_POSITIVE;
    }         

    // --- set by operator=
    DvEvent& operator=( const DvEvent& tint )
    {
        set(tint.start(),tint.end());
        return *this; 
    }

    // --- set start and end; default option is NO_TIME_SENSE i.e. store
    // start as _t1, end as _t2 regardless of whether _t1>_t2 or _t2<_t1 
    // TIME_SENSE_NEGATIVE: force _t1>=_t2
    // TIME_SENSE_POSITIVE: force _t1<=_t2
    void set( const DvTime& st, const DvTime& ed, Sense option=NO_TIME_SENSE)
    {
        switch(option)
        {
            case NO_TIME_SENSE:
                _t1=st;
                _t2=ed;
                break;
            case TIME_SENSE_NEGATIVE:
                 _t1 = (_t1>_t2) ? _t1:_t2;
                 _t2 = (_t1<_t2) ? _t1:_t2;
                break;
            case TIME_SENSE_POSITIVE:
                 _t1 = (_t1<_t2) ? _t1:_t2;
                 _t2 = (_t1>_t2) ? _t1:_t2;
         }
     }  
	
	bool setFromISOstring(const char* ISOstring)
	{
		string ISO_s1(ISOstring);
		if(ISO_s1.length() < 21) return false;
		
		string ISO_s2(ISOstring);
		if(ISO_s2.length() < 21) return false;
		
		size_t ptr = ISO_s1.find('/');
		if(ptr == string::npos) return false;
		
		ISO_s1 = ISO_s1.substr(0, ptr);
		ISO_s2.replace(0, ptr+1, "");
		
		if( !_t1.setFromISOstring(ISO_s1.c_str()) ) return false;
		if( !_t2.setFromISOstring(ISO_s2.c_str()) ) return false; 
		
		return true;
	}

	bool setFromISOstring(string ISOstring)
	{
		return setFromISOstring(ISOstring.c_str());
	}
	    
    // --- force ordering of start and end to be positive
    void force_sense( Sense option=TIME_SENSE_POSITIVE )
    {
         switch(option)
         {
             case TIME_SENSE_POSITIVE:
             if (_t1>_t2)
                 reverse();
             break;
             case TIME_SENSE_NEGATIVE:
             if (_t1 < _t2)
                 reverse();
             break;
             case NO_TIME_SENSE:
             break;
         }
    }
    
    // --- set start time (_t1) 
    void set_start( const DvTime& st)
    {     
        _t1=st;
    }     
    
     
    //  --- set end time (_t2)
	void set_end( const DvTime& ed)
    {
		_t2=ed;
    }     
    
    // --- reverse start and end
    void reverse()
    {
        DvTime ttemp=_t1;
        _t1=_t2;
        _t2=ttemp;
    }

    // shift interval by given amount of time
    DvEvent& operator+=(double offset)
    {
        _t1.time_nsec += offset;
        _t2.time_nsec += offset;
        return *this;
    }
        
    DvEvent& operator-=(double offset)
    {
        _t1.time_nsec -= offset;
        _t2.time_nsec -= offset;
        return *this;
    }

    // --- Set Operations
    DvEvent event_intersect(const DvEvent& ti) const;
    DvEvent event_union(const DvEvent& ti) const;

    bool contains( const DvTime& t ) const;
    bool intersects( const DvEvent& tivl ) const;
    bool empty() const { return durationNano() <= 1 ;}

    double operator*(const double multiplier) const
    {
		double duration = duration_abs();
        return duration * multiplier;
    }
    
    // --- Logical Operations
    bool operator==(const DvEvent& ti) const
    {
        return (start()==ti.start() && end()==ti.end());
    }
    
    bool operator!=(const DvEvent& ti) const
    {
        return !(*this==ti);
    }
    
	bool operator>(const DvEvent& ti) const
    {
		if( start() > ti.start() ) return true;
	
		if( start() == ti.start() && end() > ti.end() ) return true;
		
		return false;

    }

};


// Utilities as DvTime is not just a double

inline DvTime max(DvTime &t1, DvTime &t2)
{
	if (t1 > t2) return t1;
	else return t2;
}
inline DvTime min(DvTime &t1, DvTime &t2)
{
	if (t1 < t2) return t1;
	else return t2;
}


#endif // #ifndef DV_TIME_H



